# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.


from pytest import approx

from decaylanguage.decay.decay import DaughtersDict
from decaylanguage.decay.decay import DecayMode
from decaylanguage.decay.decay import DecayChain


def test_DaughtersDict_constructor_from_dict():
    dd = DaughtersDict({'K+': 1, 'K-': 2, 'pi+': 1, 'pi0': 1})
    assert dd == {'K+': 1, 'K-': 2, 'pi+': 1, 'pi0': 1}


def test_DaughtersDict_constructor_from_list():
    dd = DaughtersDict(['K+', 'K-', 'K-', 'pi+', 'pi0'])
    assert dd == {'K+': 1, 'K-': 2, 'pi+': 1, 'pi0': 1}


def test_DaughtersDict_constructor_from_string():
    dd = DaughtersDict('K+ K- pi0')
    assert dd == {'K+': 1, 'K-': 1, 'pi0': 1}


def test_DaughtersDict_string_repr():
    dd = DaughtersDict(['K+', 'K-', 'K-', 'pi+', 'pi0'])
    assert dd.__str__() == "<DaughtersDict: ['K+', 'K-', 'K-', 'pi+', 'pi0']>"


def test_DaughtersDict_len():
    dd = DaughtersDict({'K+': 1, 'K-': 3, 'pi0': 1})
    assert len(dd) == 5


def test_DaughtersDict_add():
    dd1 = DaughtersDict({'K+': 1, 'K-': 2, 'pi0': 3})
    dd2 = DaughtersDict({'K+': 1, 'K-': 1})
    dd3 = dd1 + dd2
    assert len(dd3) == 8


def test_DaughtersDict_to_string():
    dd1 = DaughtersDict({'K+': 1, 'K-': 2, 'pi0': 3})
    assert dd1.to_string() == 'K+ K- K- pi0 pi0 pi0'


def test_DecayMode_constructor_default():
    dm = DecayMode()
    assert dm.bf == 0
    assert dm.daughters == DaughtersDict()
    assert dm.metadata == dict(model='', model_params='')


def test_DecayMode_constructor_simplest():
    dm = DecayMode(0.1234, 'K+ K-')
    assert dm.bf == 0.1234
    assert dm.daughters == DaughtersDict('K+ K-')
    assert dm.metadata == dict(model='', model_params='')


def test_DecayMode_constructor_simple():
    dd = DaughtersDict('K+ K-')
    dm = DecayMode(0.1234, dd)
    assert dm.bf == 0.1234
    assert dm.daughters == DaughtersDict('K+ K-')
    assert dm.metadata == dict(model='', model_params='')


def test_DecayMode_constructor_from_pdgids():
    dm = DecayMode.from_pdgids(0.5, [321, -321],
                               model='TAUHADNU',
                               model_params=[-0.108, 0.775, 0.149, 1.364, 0.400])
    assert dm.daughters == DaughtersDict('K+ K-')


def test_DecayMode_constructor_with_model_info():
    dd = DaughtersDict('pi- pi0 nu_tau')
    dm = DecayMode(0.2551, dd,
                   model='TAUHADNU',
                   model_params=[-0.108, 0.775, 0.149, 1.364, 0.400])
    assert dm.metadata == {'model': 'TAUHADNU',
                           'model_params': [-0.108, 0.775, 0.149, 1.364, 0.4]}


def test_DecayMode_constructor_with_user_model_info():
    dd = DaughtersDict('K+ K-')
    dm = DecayMode(0.5, dd, model='PHSP', study='toy', year=2019)
    assert dm.metadata == {'model': 'PHSP',
                           'model_params': '',
                           'study': 'toy',
                           'year': 2019}


def test_DecayMode_constructor_from_dict():
    dm = DecayMode.from_dict({'bf': 0.98823,
                              'fs': ['gamma', 'gamma'],
                              'model': 'PHSP',
                              'model_params': ''})
    assert str(dm) == "<DecayMode: daughters=gamma gamma, BF=0.98823>"


def test_DecayMode_describe_simple():
    dd = DaughtersDict('pi- pi0 nu_tau')
    dm = DecayMode(0.2551, dd, model='TAUHADNU', model_params=[-0.108, 0.775, 0.149, 1.364, 0.400])
    assert 'BF: 0.2551' in dm.describe()
    assert 'Decay model: TAUHADNU [-0.108, 0.775, 0.149, 1.364, 0.4]' in dm.describe()


def test_DecayMode_describe_with_extra_info():
    dd = DaughtersDict('K+ K-')
    dm = DecayMode(1.e-6, dd, model='PHSP', study='toy', year=2019)
    assert 'Extra info:' in dm.describe()
    assert 'study: toy' in dm.describe()
    assert 'year: 2019' in dm.describe()


def test_DecayMode_charge_conjugate():
    dd = DaughtersDict('pi- pi0 nu_tau')
    dm = DecayMode(0.2551, dd, model='TAUHADNU', model_params=[-0.108, 0.775, 0.149, 1.364, 0.400])
    dm_cc = dm.charge_conjugate()
    assert dm_cc.daughters == DaughtersDict('pi+ pi0 anti-nu_tau')
    assert 'BF: 0.2551' in dm.describe()
    assert 'Decay model: TAUHADNU [-0.108, 0.775, 0.149, 1.364, 0.4]' in dm.describe()

    dd = DaughtersDict('pi- pi0 nu(tau)')
    dd.charge_conjugate(pdg_name=True) == DaughtersDict('pi+ pi0 nu(tau)~')


def test_DecayMode_string_repr():
    dd = DaughtersDict('p p~ K+ pi-')
    dm = DecayMode(1.e-6, dd, model='PHSP')
    assert str(dm) == "<DecayMode: daughters=K+ p pi- p~, BF=1e-06>"


def test_DecayMode_number_of_final_states():
    dd = DaughtersDict('p p~ K+ pi-')
    dm = DecayMode(1.e-6, dd, model='PHSP')
    assert len(dm) == 4


dm1 = DecayMode(0.0124, 'K_S0 pi0', model='PHSP')
dm2 = DecayMode(0.692, 'pi+ pi-')
dm3 = DecayMode(0.98823, 'gamma gamma')
dc = DecayChain('D0', {'D0':dm1, 'K_S0':dm2, 'pi0':dm3})


dm1 = DecayMode(0.6770, 'D0 pi+')
dm2 = DecayMode(0.0124, 'K_S0 pi0')
dm3 = DecayMode(0.692, 'pi+ pi-')
dm4 = DecayMode(0.98823, 'gamma gamma')
dc2 = DecayChain('D*+', {'D*+':dm1, 'D0':dm2, 'K_S0':dm3, 'pi0':dm4})


def test_DecayChain_constructor_subdecays():
    assert len(dc.decays) == 3
    assert dc.mother == 'D0'


def test_DecayChain_constructor_from_dict():
    dc_dict = {'D0': [{'bf': 0.0124,
                       'fs': [{'K_S0': [{'bf': 0.692,
                                         'fs': ['pi+', 'pi-'],
                                         'model': '',
                                         'model_params': ''}]},
                              {'pi0': [{'bf': 0.98823,
                                        'fs': ['gamma', 'gamma'],
                                        'model': '',
                                        'model_params': ''}]}],
                        'model': 'PHSP',
                        'model_params': ''}]
                }
    assert DecayChain.from_dict(dc_dict).to_dict() == dc_dict


def test_DecayChain_to_dict():
    assert dc2.to_dict() == {'D*+': [{'bf': 0.677,
                        'fs': [{'D0': [{'bf': 0.0124,
                            'fs': [{'K_S0': [{'bf': 0.692,
                                'fs': ['pi+', 'pi-'],
                                'model': '',
                                'model_params': ''}]},
                                   {'pi0': [{'bf': 0.98823,
                                   'fs': ['gamma', 'gamma'],
                                   'model': '',
                                   'model_params': ''}]}],
                            'model': '',
                            'model_params': ''}]},
                            'pi+'],
                        'model': '',
                        'model_params': ''}]}


def test_DecayChain_properties():
    assert dc.bf == 0.0124
    assert dc.visible_bf == approx(0.008479803984)


def test_DecayChain_flatten():
    dc_flatten = dc2.flatten()
    assert dc_flatten.mother == dc2.mother
    assert dc_flatten.bf == dc_flatten.visible_bf


def test_DecayChain_string_repr():
    assert str(dc) == "<DecayChain: D0 -> K_S0 pi0 (2 sub-decays), BF=0.0124>"



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from __future__ import absolute_import
from __future__ import print_function

import io
import sys
import os

from setuptools import find_packages
from setuptools import setup


PYTHON_REQUIRES = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*"

INSTALL_REQUIRES = [
    'attrs>=19.2',
    'plumbum>=1.6.9',
    'numpy>=1.12',
    'pandas>=0.22',
    'six>=1.11',
    'lark-parser>=0.8.0, <0.8.6',
    'pathlib2>=2.3; python_version<"3.5"',
    'enum34>=1.1; python_version<"3.4"',
    'importlib_resources>=1.0; python_version<"3.7"',
    'cachetools; python_version<"3.3"',
    'particle==0.11.*',
    'pydot'
]

def read(*names, **kwargs):
    return io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()

def proc_readme(text):
    text = text.split('<!-- break -->')[-1]
    return '''
    <a href="https://decaylanguage.readthedocs.io/en/latest/"><img align="left" src="https://raw.githubusercontent.com/scikit-hep/decaylanguage/master/images/DecayLanguage.png"></img></a><br>

    ''' + text

def get_version():
    g = {}
    exec(open(os.path.join("decaylanguage", "_version.py")).read(), g)
    return g["__version__"]

extras = {
    'test': ['pytest'],
    'notebook': ['graphviz'],
}

setup(
    name = 'DecayLanguage',
    author = 'Henry Fredrick Schreiner III, Eduardo Rodrigues',
    author_email = 'henry.schreiner@cern.ch, eduardo.rodrigues@cern.ch',
    maintainer = 'The Scikit-HEP admins',
    maintainer_email = 'scikit-hep-admins@googlegroups.com',
    version = get_version(),
    license = 'BSD 3-Clause License',
    description = 'A language to describe particle decays, and tools to work with them.',
    long_description = proc_readme(read('README.md')) + '\n\n' + read('CHANGELOG.md'),
    long_description_content_type = "text/markdown",
    url = 'https://github.com/scikit-hep/decaylanguage',
    packages = find_packages(),
    package_data = {'': ['data/*.*']},
    python_requires = PYTHON_REQUIRES,
    install_requires = INSTALL_REQUIRES,
    tests_require = extras['test'],
    extras_require = extras,
    keywords = [
        'HEP', 'particle', 'decay', 'representation'
    ],
    classifiers = [
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
    ],
    platforms = "Any",
)



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

'''
A class representing a set of decays. Can be subclassed to provide custom converters.
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
from copy import copy
from enum import Enum
from itertools import combinations
from itertools import product

import attr
import numpy as np
import pandas as pd
import re

from lark import Lark

from particle import Particle

import os

from .decay import ModelDecay
from ..utils import filter_lines
from ..utils import split

from ..data import open_text
from .. import data

from .ampgentransform import AmpGenTransformer, get_from_parser

class LS(Enum):
    'Line shapes supported (currently)'
    RBW = 0
    GSpline = 1
    kMatrix = 2
    FOCUS = 3


@attr.s(slots=True)
class AmplitudeChain(ModelDecay):
    'This is a chain of decays (a "line")'

    lineshape = attr.ib(None)
    spinfactor = attr.ib(None)
    amp = attr.ib(1+0j, cmp=False, validator=attr.validators.instance_of(complex))
    err = attr.ib(0+0j, cmp=False, validator=attr.validators.instance_of(complex))
    fix = attr.ib(True, cmp=False, validator=attr.validators.instance_of(bool))
    name = attr.ib(None)


    # Class members keep track of additions
    all_particles = set()
    final_particles = set()

    # This is set class-wide, and only used when a line is made
    cartesian = False

    @classmethod
    def from_matched_line(cls, mat):
        '''
        This operates on an already-matched line.

        :param mat: The groupdict output of a match
        :return: A new amplitude chain instance
        '''

        getall = 'all' if hasattr(Particle, 'all') else 'table' # Support 0.4.4

        # Check to see if new particles loaded; if not, load them.
        if 998100 not in getattr(Particle, getall)():
            data_dir = os.path.dirname(os.path.realpath(__file__))
            special_filename = os.path.join(data_dir, '..', 'data', 'MintDalitzSpecialParticles.csv')
            Particle.load_table(special_filename, append=True)

        try:
            mat['particle'] = Particle.from_string(mat['name'])
        except:
            print("Failed to find particle", mat['name'], "with parsed dictionary:", mat, file=sys.stderr)
            raise


        if mat['particle'] not in cls.all_particles:
            cls.all_particles |= {mat['particle']}

        if mat['daughters']:
            mat['daughters'] = [cls.from_matched_line(d) for d in mat['daughters']]

        # if master line only
        if 'amp' in mat and not cls.cartesian:
            A = mat['amp'].real
            dA = mat['err'].real
            theta = mat['amp'].imag
            dtheta = mat['err'].imag

            mat['amp'] = A * np.exp(theta*1j)

            mat['err'] = ((dA*np.cos(theta) + A*np.sin(dtheta))
                          + (dA*np.sin(theta) + A*np.cos(dtheta))*1j)


        return cls(**mat)

    def expand_lines(self, linelist):
        '''
        Take a DecayTree -> list of DecayTrees with each dead-end daughter
        expanded to every possible combination. (recursive)

        '''

        # If the Tree has daugthers, run on those
        if self.daughters:
            dlist = [d.expand_lines(linelist) for d in self.daughters]
            retlist = []
            for ds in product(*dlist):
                newd = copy(self)
                newd.daughters = ds
                retlist.append(newd)
            return retlist

        # If the tree ends
        new_trees = [l for line in linelist if line.name ==
                     self.name for l in line.expand_lines(linelist)]
        if new_trees:
            return new_trees
        else:
            self.__class__.final_particles |= {self.particle}
            return [self]

    @property
    def ls_enum(self):
        if not self.lineshape:
            return LS.RBW
        elif 'GSpline.EFF' == self.lineshape:
            return LS.GSpline
        elif self.lineshape.startswith('kMatrix'):
            return LS.kMatrix
        elif self.lineshape.startswith('FOCUS'):
            return LS.FOCUS
        else:
            raise RuntimeError("Unimplemented lineshape {0}".format(self.lineshape))

    @property
    def full_amp(self):
        amp = self.amp
        for d in self.daughters:
            amp *= d.full_amp
        return amp


    def L_range(self, conserveParity=False):
        S = self.particle.J
        s1 = self[0].particle.J
        s2 = self[1].particle.J
        min_spin = abs(S-s1-s2)
        min_spin = min(abs(S+s1-s2), min_spin)
        min_spin = min(abs(S-s1+s2), min_spin)
        max_spin = S + s1 + s2
        if not conserveParity:
            return (min_spin, max_spin)
        else:
            raise RuntimeError("Strong decays not implemented yet")

    @property
    def L(self):
        if self.spinfactor:
            return 'S P D F'.split().index(self.spinfactor)
        min_L, _ = self.L_range()
        return min_L  # Ground state unless specified


    def _get_html(self):
        name = self.particle.html_name
        name = re.sub(r'<SPAN STYLE="text-decoration:overline">(.*)</SPAN>', u'\\1\u0305', name)

        if self.spinfactor or self.lineshape:
            name += '<br/><br/>'
        if self.spinfactor:
            name += '<font color="blue">[' + self.spinfactor + ']</font>'
        if self.lineshape:
            name += '<font color="red">[' + self.lineshape + ']</font>'
        return name

    def __str__(self):
        name = str(self.particle)
        if self.lineshape and self.spinfactor:
            name += '[' + self.spinfactor + ';' + self.lineshape + ']'
        elif self.lineshape:
            name += '[' + self.lineshape + ']'
        elif self.spinfactor:
            name += '[' + self.spinfactor + ']'
        if self.daughters:
            name += '{'+','.join(map(str, self.daughters))+'}'
        return name

    @classmethod
    def read_ampgen(cls, filename=None, text=None, grammar=None, parser='lalr', **kargs):
        '''
        Read in an ampgen file

        :param filename: Filename to read
        :param text: Text to read (use instead of filename)
        :return: array of AmplitudeChains, parameters, constants, event type
        '''

        if grammar is None:
            grammar = open_text(data, 'ampgen.lark')

        # Read the file in, ignore empty lines and comments
        if filename is not None:
            with open(filename) as f:
                text = f.read()
        elif text is None:
            raise RuntimeError("Must have filename or text")

        lark = Lark(grammar, parser=parser, transformer=AmpGenTransformer(), **kargs)
        parsed = lark.parse(text)

        event_type, = get_from_parser(parsed, 'event_type')

        invert_lines = get_from_parser(parsed, 'invert_line')
        cplx_decay_lines = get_from_parser(parsed, 'cplx_decay_line')
        cart_decay_lines = get_from_parser(parsed, 'cart_decay_line')
        variables = get_from_parser(parsed, 'variable')
        constants = get_from_parser(parsed, 'constant')

        try:
            all_states = [Particle.from_string(n) for n in event_type]
        except:
            print("Did not find at least one of the state particles from", *event_type)
            raise

        fcs = get_from_parser(parsed, 'fast_coherent_sum')
        if fcs:
            fcs, = fcs
            fcs, = fcs.children
            cls.cartesian = bool(fcs)


        # TODO: re-enable this
        # Combine dual line Cartesian lines into traditional cartesian lines
        # for a, b in combinations(cart_decay_lines, 2):
        #     if a['name'] == b['name']:
        #        if a['cart'] == 'Re' and b['cart'] == 'Im':
        #            pass
        #        elif a['cart'] == 'Im' and b['cart'] == 'Re':
        #            a, b = b, a
        #        else:
        #            raise RuntimeError("Can't process a line with *both* components Re or Im")
        #        new_string = "{a[name]} {a[fix]} {a[amp]} {a[err]} {b[fix]} {b[amp]} {b[err]}".format(
        #            a=a, b=b)
        #        real_lines.append(ampline.dual.match(new_string).groupdict())

        # Make the partial lines and constants as dataframes
        parameters = pd.DataFrame(variables,
                                  columns='name fix value error'.split()).set_index('name')

        constants = pd.DataFrame(constants,
                                 columns='name value'.split()).set_index('name')

        # Convert the matches into AmplitudeChains
        line_arr = [cls.from_matched_line(c) for c in cplx_decay_lines]

        # Expand partial lines into complete lines
        new_line_arr = [l for line in line_arr if line.particle == all_states[0]
                        for l in line.expand_lines(line_arr)]

        # Return
        return new_line_arr, parameters, constants, all_states



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class LineFailure(RuntimeError):
    def __init__(self, line, message, *args, **kwargs):
        super(LineFailure, self).__init__("{0}: {1}".format(line, message), *args, **kwargs)
